(ns com.swir.linux.tm.luarpc.core2
  (:use aleph.tcp lamina.core gloss.core gloss.io 
    [gloss.core.protocols :exclude (sizeof)]))

(defmacro recur-codec [codec]
  `(reify
     gloss.core.protocols.Reader
     (gloss.core.protocols/read-bytes [this#  buf-seq#]
       (gloss.core.protocols/read-bytes ~codec buf-seq#))
     gloss.core.protocols.Writer
     (gloss.core.protocols/sizeof [this#] nil)
     (gloss.core.protocols/write-bytes [this# buf# val#]
       (gloss.core.protocols/write-bytes ~codec buf# val#))))

(defn si2us
  "Convert a java int (signed) to an unsigned short stored in a Java short (signed)" 
  [i]
  (if (> i 32767)
    (-> i bit-not (bit-and 0xFFFF) inc - byte)
    (byte i)))

(defn us2si 
  "Convert an unsigned short stored in a java short (signed) to a java int (signed)" 
  [b]
  (if (neg? b)
    (-> b - dec bit-not (bit-and 0xFFFF) int)
    (int b)))

(defrecord reference [i])
(def ref? (partial instance? reference))

(defrecord function [])
(def func? (partial instance? function))

(defn pre-encode-data [v]
  (cond
    (nil? v)
    [:nil]
    (instance? Boolean v)
    [:boolean (if v (byte 1) (byte 0))]
    (integer? v)
    [:integer v]
    (number? v)
    [:number v]
    (string? v)
    [:string v]
    (keyword? v)
    [:string (name v)]  
    (sequential? v)
    [:table (map-indexed #(vector (inc %1) %2) v)]
    (map? v)
    [:table (into [] v)]))

(defn post-decode-data [[t v]]
  (case t
    :nil
    nil
    :boolean
    (if (zero? v) false true)
    :integer
    (int v)
    :number
    (let [d v   i (int d)] (if (= i d) i d))
    :string
    (str v)
    :table
    (into [] v) 
    :ref
    (reference. (us2si (dec v)))))

(declare deref-elem) 

(defn- deref-tab [e b]
  (let [idx (count b)
        prom (promise)
        nb (conj b prom)
        [flat nb] (reduce #(let [[a b] %1 [ne nb] (deref-elem %2 b)] [(conj a ne) nb]) [[] nb] (mapcat identity e))
        m (apply array-map flat)
        c (take (count (keys m)) (rest (range)))
        ne (if (= c (keys m)) (into [] (vals m)) m)]
    (deliver prom ne)
    [ne (assoc nb idx ne)]))

(defn deref-elem
  ([e] (first (deref-elem e [])))
  ([e b]
    (cond
      (string? e) [e (conj b e)]
      (ref? e) [(get b (:i e)) b]
      (vector? e) (deref-tab e b) 
      :default [e b]))) 



(declare data)
(defcodec data-type (enum :byte {:nil 0 :number 3 :boolean 1 :string 4 :table 5 :function 6 :integer 7 :ref 0x14}))
(defcodec data-ni [:nil])
(defcodec data-bo [:boolean :byte])
(defcodec data-nb [:number :float64])
(defcodec data-st [:string (finite-frame :int16 (string :ascii))])
(defcodec data-tb [:table (repeated [(recur-codec data) (recur-codec data)] :prefix :int16)])
(defcodec data-rf [:ref :int16])
(defcodec data-in [:integer :int32])
(def data (compile-frame
            (header data-type
              {:nil data-ni :boolean data-bo :number data-nb :string data-st :table data-tb :ref data-rf :integer data-in}
              first)
            pre-encode-data post-decode-data))
(defcodec command-type (enum :byte {:send 0, :resp 1}))
(defcodec command-send [:send :byte (finite-frame :int32 [data (repeated data :prefix :none)])])
(defcodec command-resp [:resp :byte (finite-frame :int32 [:byte (repeated data :prefix :none)])])
(defcodec command (header command-type {:send command-send :resp command-resp} first))

(def timeout 20000)

(defprotocol Handler
  (call [this params]))

(extend-type com.swir.linux.tm.luarpc.ILuaRPCHandler
  Handler
  (call [this params]
        (.call this params)))

(let [channels (atom {})
      handlers (atom {})
      channel-id (let [i (atom 0)]
                   (fn [] (swap! i inc)))
      message-id (let [i (atom (byte 0))]
                   (fn [] (swap! i #(byte (mod (inc %) 128)))))]
  
  (defn- handle [token [t i [a b]]]
    (when (= t :send)
      (when-let [handler (@handlers (str token "-" a))]
        (let [params (map deref-elem b)
              ch (@channels token)
              resp (call handler params)]
          (enqueue ch [:resp i [(byte 0) (or resp [])]])))))


  (defn- remove-handlers [handlers-map token]
    (into {} (remove (fn [[k v]] (.startsWith k token)) handlers-map)))
  
  (defn start-client [host port]
    (let [token (str host ":" port "-" (channel-id))
          ch (wait-for-result (tcp-client {:host host :port port :frame command}) timeout)]
      (receive-all ch (partial handle token))
      (swap! channels assoc token ch)
      token))
  
  (defn stop-client [token]
    (when-let [ch (@channels token)]
      (swap! handlers remove-handlers token)
      (swap! channels dissoc token)
      (close ch)))
  
  (defn send-cmd [token cmd & args]
    (let [id (message-id)
          ch1 (@channels token)
          ch2 (channel)
          ti (System/currentTimeMillis)
          handler (fn [message])]
      (siphon ch1 ch2)
      (enqueue ch1 [:send id [cmd (or args [])]])
      (loop [[t i [a b]] (wait-for-message ch2 timeout)]
        (if (and (= i id) (= t :resp))
          (do (close ch2) (map deref-elem b))
          (recur (wait-for-message ch2 (- timeout (- (System/currentTimeMillis) ti))))))))
  
  (defn add-handler [token name handler]
    (swap! handlers assoc (str token "-" name) handler))
  
  (defn remove-handler [token name]
    (swap! handlers dissoc (str token "-" name)))
  
  (defn is-connected [token]
    (if-let [ch (@channels token)] 
      (if (closed? ch) false true) 
      false))
  
)